#!/usr/bin/env bash


: ${DEBUG:="false"}
: ${ELMA365_INSTALL_MODE:="install"}
: ${ELMA365_CLUSTER_CONNECTION_STRING:=""}
: ${ELMA365_EDITION:="standard"}
: ${ELMA365_INSTALL_DIR:="/usr/local/bin"}
: ${ELMA365_SETUP_VERSION:="2022.8.4"}
: ${ELMA365_LANG:="ru"}

function usage() {
  echo "Usage: install [ -d | --debug ] 
               [ -e | --edition EDITION ]"
  exit 2
}

PARSED_ARGUMENTS=$(getopt -a -n install -o de: --long debug,edition,join-cluster,upgrade-node,lang: -- "$@")
VALID_ARGUMENTS=$?
if [ "$VALID_ARGUMENTS" != "0" ]; then
  usage
fi


#eval set -- "$PARSED_ARGUMENTS"
while [[ $# > 0 ]] ;
do
  case "$1" in
    -d | --debug) DEBUG="true" ; shift ;;
    -e | --edition) ELMA365_EDITION="$2" ; shift 2 ;;
    --lang) ELMA365_LANG="$2" ; shift 2 ;;
    --join-cluster) ELMA365_INSTALL_MODE="join-cluster"; ELMA365_CLUSTER_CONNECTION_STRING="$2"; shift 2 ;;
    --upgrade-node) ELMA365_INSTALL_MODE="upgrade-node"; shift ;;
    --) shift; break ;;
    *) echo "Unexpected option: $1 - this should not happen."
       usage ;;
  esac
done

# Set debug if desired
if [ "${DEBUG}" == "true" ]; then
  set -x
fi

unset CDPATH
export PATH="$PATH:/snap/bin"
cd "$( dirname "${BASH_SOURCE[0]}" )"

if [ "$EUID" -ne 0 ]; then 
  echo "Please run as root"
  exit 2;
fi

if [ "$ELMA365_INSTALL_MODE" != "upgrade-node" ]; then
  if microk8s kubectl -n argocd get applications.argoproj.io/elma365 > /dev/null 2>&1 ; then

    currentBin="$ELMA365_INSTALL_DIR/elma365ctl"

    [ ! -f "$currentBin" ] && { echo "ELMA365 installed, but elma365ctl not found"; exit 2; }

    previousVersion=$($currentBin version)

    semver_result=$(./semver.sh compare $ELMA365_SETUP_VERSION $previousVersion)
    [ "$semver_result" == "-1" ] && { echo "Newer version of ELMA365 is already installed"; exit 2; } 

    read -p "Found ELMA365 version $previousVersion. Do you want upgrade to $ELMA365_SETUP_VERSION version? [y/N]" -n 1 -r
    echo    # (optional) move to a new line
    if [[ ! $REPLY =~ ^[Yy]$ ]]; then
      echo "ELMA365 upgrade aborted";
      exit 0;
    fi
    ELMA365_INSTALL_MODE="upgrade"
    echo -e "\e[93mBe sure you have made data backup before update. Choose: [B]ackup now (default), [C]ontinue, [E]xit.\e[0m"
    read -r yn
    case $yn in
      [Bb]* ) elma365ctl dump --consistent; shift;;
      [Cc]* ) echo "Skipped"; shift;;
      [Ee]* ) echo "ELMA365 upgrade aborted"; exit 0;;
      * ) elma365ctl dump --consistent; shift;;
    esac
  fi
else
  MICROK8S_VERSION=$(snap list microk8s 2> /dev/null | tail --lines=+2 | awk '{print $2}')
  read -p "Found k8s version $MICROK8S_VERSION. Do you want to upgrade k8s to the latest version 1.21.x? [y/N]" -n 1 -r
  echo    # (optional) move to a new line
  if [[ ! $REPLY =~ ^[Yy]$ ]]; then
    echo "Node upgrade aborted";
    exit 0;
  fi
fi

# fail_trap is executed if an error occurs.
fail_trap() {
  result=$?
  if [ "$result" != "0" ]; then
    echo "Failed to install $PROJECT_NAME"
    echo -e "\tFor support, go to https://elma365.com"
  fi
#  cleanup
  exit $result
}

function executeScriptsInDir() {
  path=$1
  for script in "$path"/*.sh
  do
   # shellcheck disable=SC1090
   . "${script}"
  done
}

trap "fail_trap" EXIT
set -e

# Если присоединение к кластеру, то для него гораздо меньше шагов нужно
if [[ $ELMA365_INSTALL_MODE == "join-cluster" ]]; then
  cd phase1 || exit 2;
  executeScriptsInDir preinstall
  executeScriptsInDir install
  cd ../

  echo "cluster string $ELMA365_CLUSTER_CONNECTION_STRING"

  microk8s join "$ELMA365_CLUSTER_CONNECTION_STRING"

  exit 0;
fi

# Обновление microk8s на нодах в кластере, без обновления приложения 
if [[ $ELMA365_INSTALL_MODE == "upgrade-node" ]]; then
  cd phase2 || exit 2;
  executeScriptsInDir postinstall
  cd ../

  echo "The node has been updated"

  exit 0;
fi

echo "Installing ${ELMA365_EDITION} edition"

## Todo: prehooks

# Phase 1
cd phase1 || exit 2;
executeScriptsInDir preinstall
executeScriptsInDir install
executeScriptsInDir postinstall
cd ../

# Phase 2
cd phase2 || exit 2;
executeScriptsInDir preinstall
executeScriptsInDir install
executeScriptsInDir postinstall
cd ../
